import 'reflect-metadata';

import { NextFunction, Request, Response } from 'express';
import {validate} from "./validation.midleware";
const pn = require('../index')();
const hash = {};
const argsHash = {};
const dependencies = {};

export interface ControllerOptions {
    middlewares?: Array<(req: Request, res: Response, next: NextFunction) => void>;
}

export interface RouteOptions {
    middlewares?: Array<(req: Request, res: Response, next: NextFunction) => void>;
}
export function Service() {
    return (target: any) => {
        const types: Function[] = Reflect.getMetadata('design:paramtypes', target) || [];
        dependencies[target.name] = new target(...types.map(type => dependencies[type.name]));
    };
}
export function Controller(baseUrl: string = '', options: ControllerOptions = {}) {
    const middlewares = options.middlewares || [];

    return (target: any) => {
        const types = Reflect.getMetadata('design:paramtypes', target) || [];
        const classInstance = new target(...types.map(type => dependencies[type.name]));
        hash[target.name].forEach(action => {

            if (action.options && action.options.middlewares) {
                middlewares.push(...action.options.middlewares);
            }
            const argKey = target.name + '_' + action.action.name;
            pn.express[action.type](
                pn.config.express.apiPrefix + '/' + baseUrl + '/' + action.path,
                ...middlewares,
                async (req: Request, res: Response, next: NextFunction) => {
                    classInstance.projectName = req.params.projectName;
                    const args = argsHash[argKey] ? argsHash[argKey].map((arg) => arg(req, res, next)) : [];
                    await validate(args).then(async () =>{
                        const result = await classInstance[action.action.name](...args);
                        res.send(result);
                    }).catch(error => res.send(error));
                },
            );
        });
    };
}
function addMethod(path, options: RouteOptions, type, target, propertyKey: string, descriptor: PropertyDescriptor) {
    if (!hash[target.constructor.name]) {
        hash[target.constructor.name] = [];
    }
    Object.getPrototypeOf(descriptor.value)
    hash[target.constructor.name].push({
        path,
        action: target[propertyKey],
        options,
        type,
    });
}

export function Get(path: string = '', options?: RouteOptions) {
    return function(target, propertyKey: string, descriptor: PropertyDescriptor) {
        addMethod(path, options, 'get', target, propertyKey, descriptor);
    };
}
export function Post(path: string = '', options?: RouteOptions) {
    return function(target, propertyKey: string, descriptor: PropertyDescriptor) {
        addMethod(path, options, 'post', target, propertyKey, descriptor);
    };
}

export function Put(path: string = '', options?: RouteOptions) {
    return function(target, propertyKey: string, descriptor: PropertyDescriptor) {
        addMethod(path, options, 'put', target, propertyKey, descriptor);
    };
}

export function Delete(path: string = '', options?: RouteOptions) {
    return function(target, propertyKey: string, descriptor: PropertyDescriptor) {
        addMethod(path, options, 'delete', target, propertyKey, descriptor);
    };
}

function createRouteReqParamDecorator(path: string) {
    return (): ParameterDecorator => (target, key, index) => {
        const methodParamtypes:any[] = Reflect.getMetadata('design:paramtypes', target, key) || [];

        methodParamtypes.forEach((type, i) => {
            if(!type.constructor) methodParamtypes[i] = undefined
        });

        const hashKey = target.constructor.name + '_' + key.toString();
        if (!argsHash[hashKey]) {
            argsHash[hashKey] = [];
        }
        argsHash[hashKey][index] = (req, res, next) => !methodParamtypes[index] ? req[path] : new methodParamtypes[index](req[path]);
        return undefined;
    };
}

export const Body: () => ParameterDecorator = createRouteReqParamDecorator('body');
export const Param: () => ParameterDecorator = createRouteReqParamDecorator('params');
export const User: () => ParameterDecorator = createRouteReqParamDecorator('user');
