const express = require('express');
const compression = require('compression');
const bodyParser = require('body-parser');
const cookieParser = require('cookie-parser');
const helmet = require('helmet');
const glob = require('glob').sync;
const path = require('path');
const { defaults } = require('lodash');
const cors = require('cors');
import * as mime from 'mime';
module.exports = app => {
    // Skip if configuration is not provided.
    if (!app.config.express) {
        app.logger.info('Progressive Node [express]: no configuration found.');
        return;
    }

    app.config.express = defaults(app.config.express, {
        port: 9000,
        routes: 'lib/routes',
        apiPrefix: '',
        suppressRequestLogging: true,
        staticFiles: {
            publicPrefix: '',
        },
        maxBodySize: '60mb',
    });

    const corsOptions = {
        origin: app.config.corsAllowedOrigins,
        optionsSuccessStatus: 200, // some legacy browsers (IE11, various SmartTVs) choke on 204
    };

    const isAliveRoute = `${app.config.express.apiPrefix}/isalive`;

    /**
     * This function will handle all logging requests
     * @param req
     * @param res
     * @param next
     */
    function logEveryRequest(req, res, next) {
        if (!app.config.express.suppressRequestLogging && !req.originalUrl.includes('isalive')) {
            let projectName = '';
            // Getting projectName from url, The url always looks like /projects/:projectName
            if (req.originalUrl.includes('projects/')) {
                const urlArr = req.originalUrl.split('/');
                const projectsStrIndex = urlArr.findIndex(elem => elem === 'projects');
                projectName = projectsStrIndex < urlArr.length - 1 ? urlArr[projectsStrIndex + 1] : projectName;
            }
            app.logger.info(
                {
                    method_name: 'logEveryRequest',
                    project_name: projectName,
                    mobile_app_version: req.headers.mobile_app_version,
                },
                `${req.method} ${req.originalUrl}`,
            );
        }
        next();
    }

    const server = express();

    server
        .use(logEveryRequest)
        .use(compression())
        .use(cors(corsOptions)) // usage of cors and allow cross cloud client requests (from aws to gcp for example)
        .use(bodyParser.urlencoded({ extended: true }))
        .use(bodyParser.json({ limit: app.config.express.maxBodySize }))
        .use(cookieParser())
        .use(helmet());

    app.expressAsyncWrap = fn => (req, res, next) => {
        Promise.resolve(fn(req, res, next)).catch(next);
    };

    ['all', 'get', 'post', 'head', 'put', 'delete', 'use'].forEach(methodName => {
        app[methodName] = (...args) => {
            const path = args.shift();

            args = args.map(callback => app.expressAsyncWrap(callback));

            server[methodName](path, ...args);
        };
    });

    app.express = server;

    // This request is called every few seconds for health check
    app.get(isAliveRoute, (req, res) => {
        res.sendStatus(200);
    });

    // In order to get the current version
    app.get(`${app.config.express.apiPrefix}/version`, (req, res) => {
        res.send(app.version);
    });

    glob(path.join(app.root, app.config.express.routes, '**/**.{ts,js}')).forEach(filePath => require(filePath));

    // Serve static files
    const staticFilesConfig = app.config.express.staticFiles;

    if (staticFilesConfig.root) {
        // We cache static files for 21 days and then the browser cleans them
        // from the cache, it should depend on the amount of deployments.
        server.use(
            staticFilesConfig.publicPrefix || '/',
            express.static(path.join(app.root, staticFilesConfig.root), {
                maxAge: '21d',
                setHeaders(res, path) {
                    if (mime.getType(path) === 'application/json') {
                        res.setHeader('Cache-Control', 'public, max-age=0');
                    }
                },
            }),
        );
    }

    if (staticFilesConfig.fallback) {
        const fallbackFilePath = path.join(app.root, staticFilesConfig.fallback);

        server.use(staticFilesConfig.publicPrefix || '/', (req, res, next) => {
            if (req.route || req.method !== 'GET') {
                next();
                return;
            }

            res.sendFile(fallbackFilePath);
        });
    }
    // Log all found routes
    const routes = server._router.stack.reduce((acc, r) => {
        if (r.route && r.route.path) {
            acc.push(`\n\t${r.route.stack[0].method.toUpperCase()} ${r.route.path}`);
        }
        return acc;
    }, []);
    app.logger.info(`Progressive Node [express]: ${routes.length} routes found: ${routes}`);

    server.use((err, req, res, next) => {
        // TODO: maybe should be moved to validation component
        if (err.isJoi === true && err.name === 'ValidationError' && !res.headersSent) {
            app.logger.warn(
                { method_name: 'expressErrorHandler', error_message: err.message },
                `Joi validation error in ${req.method} ${req.originalUrl}`,
            );

            res.status(400).json({
                validationErrors: err.details.reduce((acc, e) => {
                    const propertyName = e.path.join('.');

                    if (acc[propertyName]) {
                        acc[propertyName].push(e.message);
                    } else {
                        acc[propertyName] = [e.message];
                    }

                    return acc;
                }, {}),
            });
            return;
        }

        // TODO: maybe should be moved to sql component
        if (err.name === 'SequelizeValidationError' && !res.headersSent) {
            app.logger.warn(
                { method_name: 'expressErrorHandler', error_message: err.message },
                `Sequelize validation error in ${req.method} ${req.originalUrl}`,
            );
            res.status(400).json({
                validationErrors: err.errors.reduce((acc, e) => {
                    if (acc[e.path]) {
                        acc[e.path].push(e.message);
                    } else {
                        acc[e.path] = [e.message];
                    }

                    return acc;
                }, {}),
            });
            return;
        }

        // TODO: maybe should be moved to errors component
        if (err instanceof app.ClientError) {
            app.logger.warn(
                { method_name: 'expressErrorHandler', err },
                `Client error in ${req.method} ${req.originalUrl}`,
            );
            res.status(err.status).json({ message: err.message });
            return;
        }

        app.logger.error({ method_name: 'expressErrorHandler', err }, `Error in ${req.method} ${req.originalUrl}`);

        if (res.headersSent) {
            next(err);
            return;
        }

        res.status(err.httpCode || 500).send('');
    });

    if (!app.config.express.bypassListen) {
        // If a client wish to control express listen by itself, it can bypass the I/S
        server.listen(app.config.express.port, () => {
            app.logger.info(
                {
                    method_name: 'ProgressiveNode::init',
                },
                `Port ${app.config.express.port}. Environment: ${app.env}. Application: ${app.name}.`,
            );
        });
    }
};
