import { mkdirSync, writeFileSync } from 'fs';
import { sync as glob } from 'glob';
import { Project, Scope, SourceFile } from 'ts-morph';
export interface Config {
    clientPath: string;
    serverPath: string;
    decorators: { Get: string; Post: string; [key: string]: string };
    httpServiceTemplate: string;
}

export const generateClientApi = (config: Config) => {
    const clientPath = config.clientPath;
    const serverPath = config.serverPath;

    mkdirSync(clientPath, { recursive: true });
    writeFileSync(clientPath + '/http.service.ts', config.httpServiceTemplate);

    const project = new Project({ tsConfigFilePath: serverPath + '/tsconfig.json' });
    const paths = glob(serverPath + '/**/*controller.ts');
    const files: SourceFile[] = [];
    paths.forEach(filePath => {
        const file = project.addSourceFileAtPath(filePath);
        const c = file.getClasses()[0];
        const basePath = c
            .getDecorator('Controller')
            .getArguments()[0]
            .compilerNode.getText()
            .replace(/[/'/"]/g, '');

        // Remove all class Decorators
        c.getDecorators().forEach(d => d.remove());

        // Remove class constrctors & add one with di injectable
        c.getConstructors().forEach(constructor => constructor.remove());
        c.addConstructor({
            parameters: [
                {
                    isReadonly: true,
                    type: 'APIService',
                    name: 'api',
                    scope: Scope.Private,
                },
            ],
        });

        // Add necessary imports
        file.addImportDeclaration({
            namedImports: ['APIService'],
            moduleSpecifier: './http.service',
        });

        const methods = c.getMethods();
        methods.forEach(method => {
            const type = (method.getReturnTypeNode() || method.getReturnType())
                .getText(method)
                .replace('Promise<', '')
                .replace('>', '');
            let resolver = 'resolve(data)';
            let replacment = '';
            if (type !== 'any' && !type.includes('{') && !['string', 'number', 'date'].some(l => l == type)) {
                const isArray = type.includes('[]');
                if (isArray) {
                    const arrayType = type.replace('[]', '');
                    resolver = `resolve(data.map(d => new ${arrayType}(d)))`;
                } else {
                    resolver = `resolve(new ${type}(data))`;
                }
            }

            method.setReturnType(`Promise<${type}>`);

            method.getDecorators().forEach(d => {
                const name = d.getName();
                if (!config.decorators[name]) {
                    return d.remove();
                }
                const args = d.getArguments();
                const methodPath = args[0]
                    ? args[0].compilerNode.getText().replace(/['"]/g, '')
                    : // .split(':')[0]
                      '';
                const body = method
                    .getParameters()
                    .filter(p => p.getDecorators().find(d => d.getName() === 'Body'))
                    .map(p => p.compilerNode.name.getText())
                    .join(', ');
                const params =
                    method
                        .getParameters()
                        .filter(p => p.getDecorators().find(d => d.getName() === 'Param'))
                        .map(p =>
                            p.compilerNode.name
                                .getText()
                                .replace(/[/{/}' ']/g, '')
                                .split(','),
                        )[0] || [];
                let path = basePath + (methodPath ? '/' + methodPath : '');
                path = path.replace(':projectName', '').replace('//', '/')
                params.forEach(param => {
                    //'`' + projectName + '`/eventTypes'
                    ///:projectName/eventTypes
                    path = path.replace(':' + param, `'+ ${param} + '`);
                });
                replacment = config.decorators[name].replace('{url}', path).replace('{body}', body ? ', ' + body : '');
                d.remove();
            });

            method.getParameters().forEach(p => {
                const bodyDecorator = p.getDecorators().find(d => d.getName() === 'Body' || d.getName() === 'Param');
                if (!bodyDecorator) {
                    return p.remove();
                }

                p.getDecorators().forEach(d => d.remove());
            });

            const implementation = method.getImplementation();

            replacment = replacment.replace('{resolve}', resolver);
            implementation.setBodyText(replacment);
        });

        file.fixMissingImports()
            .organizeImports()
            .formatText();

        files.push(file);
    });
    return files;
};
