import { Db } from 'mongodb';
import { AUTH_ERRROS, AuthError } from 'progressive-node/components/authentication/AuthError';
import { Sequelize } from 'sequelize';

import {Logger} from "./types/logger";

const util = require('util');
const path = require('path');
import {Router} from "express";
const components = {
    config: require('progressive-node/components/config'),
    logger: require('progressive-node/components/logger'),
    monitor: require('progressive-node/components/monitor'),
    express: require('./express/express'),
    authentication: require('progressive-node/components/authentication'),
    cloud: require('progressive-node/components/cloud'),
    cache: require('progressive-node/components/cache'),
    errors: require('progressive-node/components/errors'),
    validation: require('progressive-node/components/validation'),
    mongo: require('progressive-node/components/mongo'),
    sql: require('progressive-node/components/sql'),
    graphics: require('progressive-node/components/graphics'),
    redis: require('progressive-node/components/redis'),
    graphql: require('progressive-node/components/graphql'),
    documentation: require('progressive-node/components/documentation'),
};
const componentsOrder = [
    'config',
    'logger',
    'monitor',
    'cloud',
    'errors',
    'redis',
    'cache',
    'mongo',
    'sql',
    'authentication',
    'validation',
    'graphics',
    'express',
    'documentation',
    'graphql',
];

const applications = Object.create(null);

class ProgressiveNode {
    AUTH_ERRORS: typeof AUTH_ERRROS;
    AuthError: typeof AuthError;
    ClientError: any;
    Mahsan: any;
    Sequelize: typeof Sequelize;
    validate: (field: string, schema: any) => any;
    logger: Logger =  components.logger;
    config: any;
    monitor: any;
    cloud: any;
    errors: any;
    cache: any;
    authentication: any;
    validation: any;
    graphics: any;
    express: Router;
    graphql: any;
    db: {
        redis: any;
        sql: any;
        mongo: (dbName: string) => Promise<Db>;
        sqlQuerySingleConnection: Function;
    };


    private shutdownHandlers = [];
    private name: string;
    private components;
    private componentsOrder;
    private version: string;
    private root;
    private env;
    constructor(name) {
        this.name = name || 'progressive-node';

        // Clone components information.
        this.components = { ...components };
        this.componentsOrder = componentsOrder.slice();
    }

    init(root, env): ProgressiveNode {
        this.env = env;
        this.root = root;
        this.version = '0.0.0';

        let packageError = null;

        try {
            const packageJson = require(path.join(this.root, 'package.json'));
            this.version = packageJson.version;
        } catch (err) {
            // We can't show the error here because logger not loaded yet.
            packageError = err;
        }

        // Init components.
        this.componentsOrder.forEach(componentName => this.components[componentName](this));

        if (packageError) {
            this.logger.warn(packageError, "Can't read package.json file.");
        }

        // If you running multiple applications in one process, will log
        // exception for all applications (same for promises).
        process.on('uncaughtException', err => {
            this.logger.fatal({ method_name: 'uncaughtException', err }, 'uncaughtException');
        });
        process.on('unhandledRejection', err => {
            this.logger.fatal({ method_name: 'unhandledRejection', err }, 'unhandledRejection');
        });

        // Shutdown handlers.
        const shutdown = () => {
            Promise.all(this.shutdownHandlers.map(fn => Promise.resolve(fn())))
                .then(() => process.exit(0))
                .catch(err => {
                    this.logger.fatal({ method_name: 'shutdownHandlerException', err }, 'shutdownHandlerException');
                    process.exit(1);
                });
        };

        process.on('SIGTERM', shutdown);
        process.on('SIGINT', shutdown);
        return this;
    }

    onShutdown(fn) {
        this.shutdownHandlers.push(fn);
    }

    promisifyAll(obj, skip = []) {
        Object.keys(obj).forEach(key => {
            if (!skip.includes(key)) {
                if (typeof obj[key] === 'function') {
                    obj[key] = util.promisify(obj[key]);
                }
            }
        });
    }
}

export = (applicationName?: string): ProgressiveNode => {
    const appNames = applications ? Object.keys(applications) : [];
    if ((!applicationName || typeof applicationName !== 'string') && !appNames.length) {
        throw Error('You should provide application name.');
    }

    if (!applications[applicationName] && applicationName) {
        applications[applicationName] = new ProgressiveNode(applicationName);
    }

    return applications[applicationName || appNames[0]] as ProgressiveNode;
};
